@@ -3,7 +3,7 @@ package Crypt::JWT;
 use strict;
 use warnings;
 
-our $VERSION = '0.022';
+our $VERSION = '0.024';
 
 use Exporter 'import';
 our %EXPORT_TAGS = ( all => [qw(decode_jwt encode_jwt)] );
@@ -76,8 +76,9 @@ sub _prepare_oct_key {
 
 sub _kid_lookup {
   my ($kid, $kid_keys, $alg) = @_;
-  $kid_keys = eval { decode_json($kid_keys) } unless ref $kid_keys;
-  return undef unless ref $kid_keys eq 'HASH';
+  return undef if !defined $kid || !defined $alg;
+  $kid_keys = eval { decode_json($kid_keys) } if $kid_keys && !ref $kid_keys;
+  croak "JWT: kid_keys must be a HASHREF or a valid JSON/HASH" if ref $kid_keys ne 'HASH';
   my $found;
   if (exists $kid_keys->{keys} && ref $kid_keys->{keys} eq 'ARRAY') {
     #FORMAT: { keys => [ {kid=>'A', kty=>?, ...}, {kid=>'B', kty=>?, ...} ] }
@@ -167,42 +168,74 @@ sub _verify_claims {
   }
 
   ### iss
-  if(exists $payload->{iss}) {
-    if (ref $args{verify_iss} eq 'Regexp') {
-      croak "JWT: iss claim re check failed" unless $payload->{iss} =~ $args{verify_iss};
+  if (exists $args{verify_iss}) {
+    if (exists $payload->{iss}) {
+      if (ref $args{verify_iss} eq 'Regexp') {
+        croak "JWT: iss claim re check failed" unless $payload->{iss} =~ $args{verify_iss};
+      }
+      elsif (ref $args{verify_iss} eq 'CODE') {
+        croak "JWT: iss claim check failed" unless $args{verify_iss}->($payload->{iss});
+      }
+      else {
+        croak "JWT: verify_iss must be Regexp or CODE";
+      }
     }
-    elsif (ref $args{verify_iss} eq 'CODE') {
-      croak "JWT: iss claim check failed" unless $args{verify_iss}->($payload->{iss});
+    elsif ($args{verify_iss}) {
+      croak "JWT: iss claim required but missing"
     }
   }
 
   ### sub
-  if(exists $payload->{sub}) {
-    if (ref $args{verify_sub} eq 'Regexp') {
-      croak "JWT: sub claim re check failed" unless $payload->{sub} =~ $args{verify_sub};
+  if (exists $args{verify_sub}) {
+    if (exists $payload->{sub}) {
+      if (ref $args{verify_sub} eq 'Regexp') {
+        croak "JWT: sub claim re check failed" unless $payload->{sub} =~ $args{verify_sub};
+      }
+      elsif (ref $args{verify_sub} eq 'CODE') {
+        croak "JWT: sub claim check failed" unless $args{verify_sub}->($payload->{sub});
+      }
+      else {
+        croak "JWT: verify_sub must be Regexp or CODE";
+      }
     }
-    elsif (ref $args{verify_sub} eq 'CODE') {
-      croak "JWT: sub claim check failed" unless $args{verify_sub}->($payload->{sub});
+    elsif ($args{verify_sub}) {
+      croak "JWT: sub claim required but missing"
     }
   }
 
   ### aud
-  if(exists $payload->{aud}) {
-    if (ref $args{verify_aud} eq 'Regexp') {
-      croak "JWT: aud claim re check failed" unless $payload->{aud} =~ $args{verify_aud};
+  if (exists $args{verify_aud}) {
+    if (exists $payload->{aud}) {
+      if (ref $args{verify_aud} eq 'Regexp') {
+        croak "JWT: aud claim re check failed" unless $payload->{aud} =~ $args{verify_aud};
+      }
+      elsif (ref $args{verify_aud} eq 'CODE') {
+        croak "JWT: aud claim check failed" unless $args{verify_aud}->($payload->{aud});
+      }
+      else {
+        croak "JWT: verify_aud must be Regexp or CODE";
+      }
     }
-    elsif (ref $args{verify_aud} eq 'CODE') {
-      croak "JWT: aud claim check failed" unless $args{verify_aud}->($payload->{aud});
+    elsif ($args{verify_aud}) {
+      croak "JWT: aud claim required but missing"
     }
   }
 
   ### jti
-  if(exists $payload->{jti}) {
-    if (ref $args{verify_jti} eq 'Regexp') {
-      croak "JWT: jti claim re check failed" unless $payload->{jti} =~ $args{verify_jti};
+  if (exists $args{verify_jti}) {
+    if (exists $payload->{jti}) {
+      if (ref $args{verify_jti} eq 'Regexp') {
+        croak "JWT: jti claim re check failed" unless $payload->{jti} =~ $args{verify_jti};
+      }
+      elsif (ref $args{verify_jti} eq 'CODE') {
+        croak "JWT: jti claim check failed" unless $args{verify_jti}->($payload->{jti});
+      }
+      else {
+        croak "JWT: verify_jti must be Regexp or CODE";
+      }
     }
-    elsif (ref $args{verify_jti} eq 'CODE') {
-      croak "JWT: jti claim check failed" unless $args{verify_jti}->($payload->{jti});
+    elsif ($args{verify_jti}) {
+      croak "JWT: jti claim required but missing"
     }
   }
 }
@@ -430,7 +463,6 @@ sub _encode_jwe {
   # prepare header
   $header->{alg} = $alg;
   $header->{enc} = $enc;
-  #REMOVED: $header->{typ} = 'JWT' if !exists $header->{typ} && $args{auto_typ};
   # key
   croak "JWE: missing 'key'" if !$args{key};
   my $key = defined $args{keypass} ? [$args{key}, $args{keypass}] : $args{key};
@@ -464,31 +496,37 @@ sub _decode_jwe {
   croak "JWE: invalid iv part"     if $b64u_iv     && !$iv;
   croak "JWE: invalid tag part"    if $b64u_tag    && !$tag;
 
-  my $key = defined $args{keypass} ? [$args{key}, $args{keypass}] : $args{key};
-  if ($header->{kid} && $args{kid_keys}) {
-    my $k = _kid_lookup($header->{kid}, $args{kid_keys}, $header->{alg});
-    $key = $k if defined $k;
+  my $key;
+  if (exists $args{key}) {
+    $key = defined $args{keypass} ? [$args{key}, $args{keypass}] : $args{key};
   }
+  elsif (exists $args{kid_keys}) {
+    # BEWARE: stricter approach since 0.023
+    # when 'kid_keys' specified it croaks if header doesn't contain 'kid' value or if 'kid' wasn't found in 'kid_keys'
+    my $k = _kid_lookup($header->{kid}, $args{kid_keys}, $header->{alg});
+    croak "JWE: kid_keys lookup failed" if !defined $k;
+    $key = $k;
+  }
+  croak "JWE: missing key" if !defined $key;
 
   my $aa = $args{accepted_alg};
   if (ref($aa) eq 'Regexp') {
-    croak "JWT: alg '$header->{alg}' does not match accepted_alg" if $header->{alg} !~ $aa;
+    croak "JWE: alg '$header->{alg}' does not match accepted_alg" if $header->{alg} !~ $aa;
   }
   elsif ($aa && (ref($aa) eq 'ARRAY' || !ref($aa))) {
     my %acca = ref $aa ? map { $_ => 1 } @$aa : ( $aa => 1 );
-    croak "JWT: alg '$header->{alg}' not in accepted_alg" if !$acca{$header->{alg}};
+    croak "JWE: alg '$header->{alg}' not in accepted_alg" if !$acca{$header->{alg}};
   }
 
   my $ae = $args{accepted_enc};
   if (ref($ae) eq 'Regexp') {
-    croak "JWT: enc '$header->{enc}' does not match accepted_enc" if $header->{enc} !~ $ae;
+    croak "JWE: enc '$header->{enc}' does not match accepted_enc" if $header->{enc} !~ $ae;
   }
   elsif ($ae && (ref($ae) eq 'ARRAY' || !ref($ae))) {
     my %acce = ref $ae ? map { $_ => 1 } @$ae : ( $ae => 1 );
-    croak "JWT: enc '$header->{enc}' not in accepted_enc" if !$acce{$header->{enc}};
+    croak "JWE: enc '$header->{enc}' not in accepted_enc" if !$acce{$header->{enc}};
   }
 
-  croak "JWE: missing 'key'" if !$key;
   $header = { %$shared_unprotected, %$unprotected, %$header }; # merge headers
   my $cek = _decrypt_jwe_cek($ecek, $key, $header);
   my $aad = defined $b64u_aad ? "$b64u_header.$b64u_aad" : $b64u_header;
@@ -573,7 +611,6 @@ sub _encode_jws {
   my $b64u_payload = encode_b64u($payload);
   # prepare header
   $header->{alg} = $alg;
-  #REMOVED: $header->{typ} = 'JWT' if !exists $header->{typ} && $args{auto_typ};
   # encode header
   my $json_header = encode_json($header);
   my $b64u_header = encode_b64u($json_header);
@@ -592,29 +629,51 @@ sub _decode_jws {
   $unprotected_header = {} if ref $unprotected_header ne 'HASH';
 
   if (!$args{ignore_signature}) {
-    my $aa = $args{accepted_alg};
-    if (ref($aa) eq 'Regexp') {
-      croak "JWT: alg '$header->{alg}' does not match accepted_alg" if $header->{alg} !~ $aa;
-    }
-    elsif ($aa && (ref($aa) eq 'ARRAY' || !ref($aa))) {
-      my %acca = ref $aa ? map { $_ => 1 } @$aa : ( $aa => 1 );
-      croak "JWT: alg '$header->{alg}' not in accepted_alg" if !$acca{$header->{alg}};
-    }
     my $alg = $header->{alg};
     croak "JWS: missing header 'alg'" unless $alg;
     croak "JWS: alg 'none' not allowed" if $alg eq 'none' && !$args{allow_none};
-    # key
-    my $key = defined $args{keypass} ? [$args{key}, $args{keypass}] : $args{key};
-    my $kid = exists $header->{kid} ? $header->{kid} : $unprotected_header->{kid};
-    if (!defined $key && defined $kid && $args{kid_keys}) {
-      my $k = _kid_lookup($kid, $args{kid_keys}, $alg);
-      $key = $k if defined $k;
+    croak "JWS: alg 'none' expects no signature" if $alg eq 'none' && defined $b64u_sig && length($b64u_sig) > 0;
+
+    my $aa = $args{accepted_alg};
+    if (ref $aa eq 'Regexp') {
+      croak "JWS: alg '$alg' does not match accepted_alg" if $alg !~ $aa;
+    }
+    elsif (ref $aa eq 'ARRAY') {
+      my %acca = map { $_ => 1 } @$aa;
+      croak "JWS: alg '$alg' not in accepted_alg" if !$acca{$alg};
+    }
+    elsif (defined $aa) {
+      croak "JWS: alg '$alg' not accepted_alg" if $aa ne $alg;
+    }
+
+    if ($alg ne 'none') {
+      my $key;
+      if (exists $args{key}) {
+        $key = defined $args{keypass} ? [$args{key}, $args{keypass}] : $args{key};
+      }
+      elsif (exists $args{kid_keys}) {
+        # BEWARE: stricter approach since 0.023
+        # when 'kid_keys' specified it croaks if header doesn't contain 'kid' value or if 'kid' wasn't found in 'kid_keys'
+        my $kid = exists $header->{kid} ? $header->{kid} : $unprotected_header->{kid};
+        my $k = _kid_lookup($kid, $args{kid_keys}, $alg);
+        croak "JWS: kid_keys lookup failed" if !defined $k;
+        $key = $k;
+      }
+      elsif ($args{key_from_jwk_header}) {
+        # BEWARE: stricter approach since 0.023
+        # - header 'jwk' is by default ignored (unless given: key_from_jwk_header => 1)
+        # - only RSA/ECDSA public keys are accepted
+        my $k = $header->{jwk};
+        croak "JWS: jwk header does not contain a key" if !defined $k || ref($k) ne 'HASH' || !defined $k->{kty};
+        croak "JWS: jwk header allowed only for RSA/ECDSA" if $alg !~ /^(RS|PS|ES)/ || $k->{kty} !~ /^(RSA|EC)$/;
+        croak "JWS: jwk header must be a public key" if $k->{d} || $k->{p} || $k->{q} || $k->{dp} || $k->{dq} || $k->{qi};
+        $key = $k;
+      }
+      croak "JWS: missing key" if !defined $key;
+
+      my $valid = _verify_jws($b64u_header, $b64u_payload, $b64u_sig, $alg, $key);
+      croak "JWS: invalid signature" if !$valid;
     }
-    # if no key given, try to use 'jwk' value from header
-    $key = $header->{jwk} if !$key && $header->{jwk};
-    croak "JWS: missing 'key'" if !$key && $alg ne 'none';
-    my $valid = _verify_jws($b64u_header, $b64u_payload, $b64u_sig, $alg, $key);
-    croak "JWS: decode failed" if !$valid;
   }
   my $payload = decode_b64u($b64u_payload);
   croak "JWS: invalid payload part" if $b64u_payload && !$payload;
@@ -788,7 +847,6 @@ Mandatory argument, a string with either JWS or JWE JSON Web Token.
 =item key
 
 A key used for token decryption (JWE) or token signature validation (JWS).
-If not given, and the token header contains a C<jwk> header value, that value will be used.
 The value depends on the C<alg> token header value.
 
  JWS alg header      key value
@@ -835,10 +893,10 @@ The value depends on the C<alg> token header value.
  ECDH-ES+A192KW      private ECC key, see ECDH-ES
  ECDH-ES+A256KW      private ECC key, see ECDH-ES
 
-Example with key from C<jwk> token header:
+Example using the key from C<jwk> token header:
 
- my $data = decode_jwt(token=>$t);
- my ($header, $data) = decode_jwt(token=>$t, decode_header=>1);
+ my $data = decode_jwt(token=>$t, key_from_jwk_header=>1);
+ my ($header, $data) = decode_jwt(token=>$t, decode_header=>1, key_from_jwk_header=>1);
 
 Examples with raw octet keys:
 
@@ -943,15 +1001,35 @@ This parametes can be either a JWK Set JSON string (see RFC7517) or a perl HASH
   };
   my $payload = decode_jwt(token=>$t, kid_keys=>$keylist);
 
-Since 0.19 we also support:
+The structure described above is used e.g. by L<https://www.googleapis.com/oauth2/v2/certs>
+
+  use Mojo::UserAgent;
+  my $ua = Mojo::UserAgent->new;
+  my $google_keys => $ua->get('https://www.googleapis.com/oauth2/v2/certs')->result->json;
+  my $payload = decode_jwt(token => $t, kid_keys => $google_keys);
+
+B<SINCE 0.019> we also support alternative structure used e.g. by L<https://www.googleapis.com/oauth2/v1/certs>:
 
   use LWP::Simple;
   my $google_certs = get('https://www.googleapis.com/oauth2/v1/certs');
   my $payload = decode_jwt(token => $t, kid_keys => $google_certs);
 
-When the token header contains 'kid' item the corresponding key is looked up in C<kid_keys> list and used for token
+When the token header contains C<kid> item the corresponding key is looked up in C<kid_keys> list and used for token
 decoding (you do not need to pass the explicit key via C<key> parameter).
 
+B<INCOMPATIBLE CHANGE in 0.023:> When C<kid_keys> is specified it croaks if token header does not contain C<kid> value or
+if C<kid> was not found in C<kid_keys>.
+
+=item key_from_jwk_header
+
+B<SINCE 0.023>
+
+C<1> - use C<jwk> header value for validating JWS signature if neither C<key> nor C<kid_keys> specified, B<BEWARE: DANGEROUS, UNSECURE!!!>
+
+C<0> (default) - ignore C<jwk> header value when validating JWS signature
+
+Keep in mind that enabling C<key_from_jwk_header> requires C<jwk> header to exist and be an valid RSA/ECDSA public key (otherwise it croaks).
+
 =item allow_none
 
 C<1> - accept JWS tokens with C<none> 'alg' header value (which means that token has no signature), B<BEWARE: DANGEROUS, UNSECURE!!!>
@@ -1016,6 +1094,9 @@ C<1> - return decoded header as a return value of decode_jwt()
 
 =item verify_iss
 
+B<INCOMPATIBLE CHANGE in 0.024:> If C<verify_iss> is specified and
+claim C<iss> (Issuer) is completely missing it is a failure since 0.024
+
 C<CODE ref> - subroutine (with 'iss' claim value passed as argument) should return C<true> otherwise verification fails
 
 C<Regexp ref> - 'iss' claim value has to match given regexp otherwise verification fails
@@ -1024,6 +1105,9 @@ C<undef> (default) - do not verify 'iss' claim
 
 =item verify_aud
 
+B<INCOMPATIBLE CHANGE in 0.024:> If C<verify_iss> is specified and
+claim C<aud> (Audience) is completely missing it is a failure since 0.024
+
 C<CODE ref> - subroutine (with 'aud' claim value passed as argument) should return C<true> otherwise verification fails
 
 C<Regexp ref> - 'aud' claim value has to match given regexp otherwise verification fails
@@ -1032,6 +1116,9 @@ C<undef> (default) - do not verify 'aud' claim
 
 =item verify_sub
 
+B<INCOMPATIBLE CHANGE in 0.024:> If C<verify_iss> is specified and
+claim C<sub> (Subject) is completely missing it is a failure since 0.024
+
 C<CODE ref> - subroutine (with 'sub' claim value passed as argument) should return C<true> otherwise verification fails
 
 C<Regexp ref> - 'sub' claim value has to match given regexp otherwise verification fails
@@ -1040,6 +1127,9 @@ C<undef> (default) - do not verify 'sub' claim
 
 =item verify_jti
 
+B<INCOMPATIBLE CHANGE in 0.024:> If C<verify_iss> is specified and
+claim C<jti> (JWT ID) is completely missing it is a failure since 0.024
+
 C<CODE ref> - subroutine (with 'jti' claim value passed as argument) should return C<true> otherwise verification fails
 
 C<Regexp ref> - 'jti' claim value has to match given regexp otherwise verification fails
